"""
Name: NCGMP1.0_md2.py
Purpose: Automates the creation of metadata records and entity-attribute
    information for a NCGMP09-compliant geologic map geodatabase

PROBLEMS:
    In ArcCatalog 9.3, updating metadata for a feature dataset clobbers detailed entity
    descriptions, detailed entity description sources, and all attribute information. 
   
    In ArcCatalog 9.3, updating metadata for an annotation feature class adds attribute nodes
    (attr) for a bunch of ESRI-defined fields without any definitions (attrdef) or definition
    sources (attrdefs). AARGH!

Strategy:
    A HUMAN MUST COMPLETE the metadata record for feature dataset GeologicMap.
    Then run script, which:
        Exports GeologicMap metadata to XML master record.
        Checks that a few key fields of master record all have non-zero-length text. If so:
            Master record is stripped of entity-attribute info.
            For each constituent element (table, feature dataset, feature class) of gdb:
                If forceRebuild = False:
                    If metadata record exists, we start with this, else start with master record.
                If forceRebuild = True:
                    Start with master record.
                    For non-spatial tables, strip spatial data organization, data quality, and
                    spatial reference framework information from master record.
                Entity-attribute information is appended/replaced.
                Metadata is restored to geodatabase.
        Records to <geodatabaseName>-metadata_omissions.txt those features and fields for which
        e-a information was not completed.

    Then A HUMAN MUST OPEN each metadata record in ArcCatalog, UPDATE the record to
    set spatial data organization and spatial reference framework information, perhaps ENTER
    DATA for features and fields identified in log file, and--as desired--export the
    metadata to .txt, .xml., or .htm files.

    Note that we were unable to figure out how to import, from a script, metadata for the
    geodatabase as a whole. So this is written to file <geodatabase.gdb>-meetadata.xml.

    Note that if you manually import metadata with automatic update, the TITLE will be
    clobbered.

    Setting forceRebuild = True is likely to clobber any corrections that have already
    been made to metadata.

    This script does not check that all required fields in master metadata record contain
    appropriate text. DO THIS YOURSELF BEFORE RUNNING THE SCRIPT. Else you will have to
    manually update, e.g., contact info, in the metadata for EACH geodatabase element.
    Or, you can fix the master record (GeologicMap) and re-run the script with
    forceRebuild = True.
            
Authors: Evan Thoms, ethoms@usgs.gov, 1/8/09
         Ralph Haugerud. rhaugerud@usgs.gov, 11/15/2010
        
Usage: Call with at least one argument, a gdb name, e.g.,

       ncgmp09_Metadata.py Camas.gdb
       
       Second and 3rd arguments set values for saveXmlFiles and
       forceRebuild, e.g.,

       ncgmp09_Metadata.py camas.gdb False True
       
       Forces saveXmlFiles = False and forceRebuild = True. Default values
       are saveXmlFiles = True and forceRebuild = False
       
Requirements: ArcGIS 9.3       
"""

import arcgisscripting
import os, os.path
import sys, copy
import codecs
from ncgmp09_definition import enumeratedValueDomainFieldList, rangeDomainDict, unrepresentableDomainDict, attribDict, entityDict
from xml.dom.minidom import *
import xml.dom.minidom

versionString = 'ncgmp09_Metadata.py  version of 15 November 2010'

ncgmp = 'NCGMP09 v1.1'
ncgmpFullRef = '"NCGMP09--draft standard format for digital publication of geologic maps, version 1.1", available at http://ngmdb.usgs.gov/Info/standards/NCGMP09/'

saveXmlFiles = False
forceRebuild = False
# flag to set ea info in geodatabase-level record to Overview description or Detailed description
masterRecordEaOverview = False  

# places where ESRI sticks "REQUIRED: ..." text, and we want to remove it
elementsRequired = ['origin','pubdate','abstract','purpose','caldate','current',
                    'progress','update','westbc','eastbc','northbc','southbc','themekt',
                    'themekey','accconst','useconst','cntorg','cntper','addrtype',
                    'city','state','postal','cntvoice']
                    

#********************************************

def addMsgAndPrint(gp,msg, severity=0): 
	# prints msg to screen and adds msg to the geoprocessor (in case this is run as a tool) 
	# print msg 
	try: 
	  for string in msg.split('\n'): 
		# Add appropriate geoprocessing message 
		if severity == 0: 
			gp.AddMessage(string) 
		elif severity == 1: 
			gp.AddWarning(string) 
		elif severity == 2: 
			gp.AddError(string) 
	except: 
		pass 

def processMD(gdbPath):
    #Master switchboard function from which all other functions are called
    gdbName = os.path.basename(gdbPath)
    
    #set globally accessible paths to the Glossary and DataSources tables
    gloss = os.path.join(gdbPath, 'Glossary')
    dataSources = os.path.join(gdbPath, 'DataSources')
    
    #get the parent folder of the geodatabase object
    gdbFolder = __parentFolder(gdbPath)
    gp.workspace = gdbPath

    #####do master record
    #export GeologicMap metadata to text file, open text file, and parse textfile to domMR
    domMR = __dumpMetadataToDom('GeologicMap',gdbFolder,1)
    
    #check domMR for text in abstract, supplemental info
    #  originator, and title fields; bail if not there
    if __keyFieldsNull(domMR):
        print
        print 'Please complete fields in master metatada record'
        print
        sys.exit()
    else:
        #open logFile
        logFile = open(gdbFolder+'/'+gdbName+'-metadata_omissions.txt', 'w')
        #purge master record of entity-attribute info
        domMR = __purgeChildren(domMR,'eainfo')
        #purge master record of spatial data organization info
        domMR = __purgeChildren(domMR,'spdoinfo')
        #purge certain elements of text that contains "REQUIRED:"
        domMR = __purgeRequired(domMR)
        #clean dross off of title
        domMR = __cleanTitle(domMR)

        # geodatabase as a whole
        domGdb = copy.deepcopy(domMR)
        tbs = gp.listtables()
        fds = gp.listdatasets('', 'feature')
        fdDict = {}
        for fd in fds:
            gp.workspace = fd
            fcs = gp.listfeatureclasses()
            if len(fcs) > 0:
                fcString = 'Contains feature classes: '
                for fc in fcs:
                    fcString = fcString + fc +', '
                fcString = fcString[:-2]
            else:
                fcString = 'Empty feature dataset.'
            if entityDict.has_key(fd):
                fcString = entityDict[fd]+' '+fcString
            else:
                if fd.find('CrossSection') > -1:
                    fcString = entityDict['CrossSection']+' '+fcString
            fdDict[fd] = fcString
            gp.workspace = gdbPath
        if masterRecordEaOverview:
            domGdb = __updateGdbDomOverview(domGdb,tbs,fds,fdDict,logFile, gdbName, gdbFolder)
        else:
            domGdb = __updateGdbDomDetailed(domGdb,tbs,fds,fdDict,logFile, gdbName, gdbFolder)
        domGdb.unlink()

        # rename master record
        newXmlName = os.path.join(gdbFolder,gdbName[:-4]+'-metadata.xml')
        if os.path.exists(newXmlName):
            os.remove(newXmlName)
        os.rename(os.path.join(gdbFolder,gdbName+'.xml'),newXmlName)

        ## non-spatial tables
        domTbM = copy.deepcopy(domMR)
        # dataquality and spatial reference elements are irrelevant for a non-spatial table
        domTbM = __purgeChildren(domTbM,'dataqual')
        domTbM = __purgeChildren(domTbM,'spref')
        for tb in gp.listtables():
            addMsgAndPrint(gp,'\tTable: '+tb)
            domTb = copy.deepcopy(domTbM)
            domTb = __updateTable(domTb,tb,gdbFolder,': table '+tb, logFile, False)
            domTb.unlink()
        domTbM.unlink()

        ## feature datasets
        for fDS in gp.listdatasets('', 'feature'):        
            addMsgAndPrint(gp,'\t'+fDS)
            gp.workspace = gdbPath
            if forceRebuild:
                domFDS = copy.deepcopy(domMR)
            else:
                domFDS = __dumpMetadataToDom(fDS,gdbFolder,1)
            domFDS = __cleanTitle(domFDS)
            domFDS = __purgeRequired(domFDS)
            domFDS = __suffixTitle(domFDS, ': feature dataset '+fDS)

            ## feature classes
            gp.workspace = fDS
            fcList = gp.listfeatureclasses()
            
            for fc in fcList:
                addMsgAndPrint(gp,'\t' * 2 + 'Feature class: '+fc)
                titleSuffix = ': feature dataset '+fDS+': feature class '+fc
                if gp.describe(fc).FeatureType == 'Annotation':
                    isAnno = True
                else:
                    isAnno = False
                domFC = __updateTable(domMR,fc,gdbFolder, titleSuffix, logFile, isAnno)
                #update the feature class element in the feature dataset metadata
                __updateFdsDom(domFDS,domFC,fc,fDS,logFile)
                domFC.unlink()
        
            __importMetadataFromDom(domFDS, fDS, gdbFolder)
            domFDS.unlink()
            
        domMR.unlink()
        logFile.close()
        

def printNumberEaChildNodes(dom):
    eaNode = dom.getElementsByTagName('eainfo')[0]
    addMsgAndPrint(gp, 'dom has '+str(len(eaNode.childNodes))+' child nodes')

def dumpDomToText(dom,fileName):
    outf = open(fileName,'w')
    dom.writexml(outf)
    outf.close()

def __purgeRequired(dom):
    for tag in elementsRequired:
        nodes = dom.getElementsByTagName(tag)
        for nd in nodes:
            # test for presence of "REQUIRED:"
            try:
                if nd.firstChild.nodeValue.find('REQUIRED:') > -1:
                    nd.removeChild(nd.firstChild)
            except:
                pass
    return dom

def __purgeChildren(dom,nodeTag):
    nodes = dom.getElementsByTagName(nodeTag)
    for aNode in nodes:
        while len(aNode.childNodes) > 0:
            aNode.removeChild(aNode.lastChild)
    return dom

def __cleanTitle(dom):
    # trim all ": table...", ":  feature..." from title
    #get the Title node for this dataset
    title = dom.getElementsByTagName('title')[0]
    titleText = title.firstChild.data
    cn = titleText.find(': feature')
    if cn > 0:
        titleText = titleText[0:cn]
    cn = titleText.find(': table')
    if cn > 0:
        titleText = titleText[0:cn]
    newTitle = dom.createElement('title')
    newTitleText = dom.createTextNode(titleText)
    newTitle.appendChild(newTitleText)
    title.firstChild.data = newTitleText
    citation = title.parentNode
    citation.replaceChild(newTitle,title)
    return dom

def __suffixTitle(dom,titleSuffix):
    # adds suffix to title text
    title = dom.getElementsByTagName('title')[0]
    titleText = title.firstChild.data
    if titleText.find(titleSuffix) == -1:  # titleSuffix isn't already present
        newTitle = dom.createElement('title')
        newTitleText = dom.createTextNode(titleText+titleSuffix)
        newTitle.appendChild(newTitleText)
        title.firstChild.data = newTitleText
        citation = title.parentNode
        citation.replaceChild(newTitle,title)
    return dom

def __dumpMetadataToDom(gdbObj,outDir,numOfTabs):
    # gdbObj is table, feature dataset, or feature class, not geodatabase!
    # folder is writable output folder
    fXML = os.path.join(outDir, gdbObj + '.xml')
    if os.path.exists(fXML):
        os.remove(fXML)
    gp.USGSMPTranslator_conversion (gdbObj, '','XML', fXML)
    addMsgAndPrint(gp,'\t' * numOfTabs+ 'Metadata for '+gdbObj+' exported to file ')
    addMsgAndPrint(gp,'\t' * numOfTabs+ fXML)
    try:
            dom = xml.dom.minidom.parse(fXML)
            # should then delete xml file
            os.remove(fXML)
    except:
        addMsgAndPrint(gp,gp.GetMessages())
        addMsgAndPrint(gp,'Failed to parse '+fXML)
    return dom   

def __keyFieldsNull(dom):
    # checks for non-zero-length text in certain fields within XML object dom
    keyFields = ('title','abstract','purpose','supplinf','northbc')
    kfNull = False
    for kf in keyFields:
        try:
            nd = dom.getElementsByTagName(kf)
            ndText = nd[0].firstChild.nodeValue
            if len(ndText) == 0 or ndText.find('REQUIRED:') > -1:
                kfNull = True
        except:
            kfNull = True
    return kfNull

def __newElement(dom,tag,text):
    nd = dom.createElement(tag)
    ndText = dom.createTextNode(text)
    nd.appendChild(ndText)
    return nd

def __updateGdbDomOverview(dom,tbs,fds,fdDict,logFile, gdbName, gdbFolder):
    # adds table names, feature dataset names, and descriptions as 
    # OVERVIEW entity-attributes elements of gdb-level metadata.
    eanode = dom.getElementsByTagName('eainfo')[0]
    for tb in tbs:
        overview = dom.createElement('overview')
        eanode.appendChild(overview)
        if entityDict.has_key(tb):
            overviewText = 'Table '+tb+': '+entityDict[tb]
        else:
            overviewText = 'Table '+tb
            logFile.write('No description for table '+tb+'\n')
        eaover = __newElement(dom,'eaover',overviewText)
        overview.appendChild(eaover)
        citationText = 'See metadata record for table '+tb+', which is part of this dataset.'
        eadetcit = __newElement(dom,'eadetcit',citationText)
        overview.appendChild(eadetcit)
        citationText = 'Also see '+ncgmpFullRef
        eadetcit = __newElement(dom,'eadetcit',citationText)
        overview.appendChild(eadetcit)    
    for fd in fds:
        overview = dom.createElement('overview')
        eanode.appendChild(overview)
        overviewText = fd+': '+fdDict[fd]
        eaover = __newElement(dom,'eaover',overviewText)
        overview.appendChild(eaover)
        citationText = 'See metadata record for feature dataset '+fd+' and its constituent feature classes, which are parts of this dataset.'
        eadetcit = __newElement(dom,'eadetcit',citationText)
        overview.appendChild(eadetcit)
        citationText = 'Also see '+ncgmpFullRef
        eadetcit = __newElement(dom,'eadetcit',citationText)
        overview.appendChild(eadetcit)       
    # __importMetadataFromDom has a trap for gdbName to avoid
    #   a failed attempt to import gdb-level metadata
    __importMetadataFromDom(dom, gdbName, gdbFolder)
    return dom

def __updateGdbDomDetailed(dom,tbs,fds,fdDict,logFile, gdbName, gdbFolder):
    # adds table names, feature dataset names, and descriptions as
    # DETAILED entity-attribute elements of gdb-level metadata. 
    eanode = dom.getElementsByTagName('eainfo')[0]
    for tb in tbs:
        detailed = dom.createElement('detailed')
        eanode.appendChild(detailed)
        enttyp = dom.createElement('enttyp')
        detailed.appendChild(enttyp)
        enttypl = __newElement(dom,'enttypl',tb)
        enttyp.appendChild(enttypl)

        if entityDict.has_key(tb):
            ndText = entityDict[tb]
        else:
            ndText = ''
            logFile.write('No description for table '+tb+'\n')
        enttypd = __newElement(dom,'enttypd',ndText)
        enttyp.appendChild(enttypd)
        enttypds = __newElement(dom,'enntypds',ncgmp)
        enttypds.appendChild(enttypds)    
    for fd in fds:
        detailed = dom.createElement('detailed')
        enttyp = dom.createElement('enttyp')
        detailed.appendChild(enttyp)
        enttypl = __newElement(dom,'enttypl',fd)
        enttyp.appendChild(enttypl)
        enttypd = __newElement(dom,'enttypd',fdDict[fd])
        enttyp.appendChild(enttypd)
        enttypds = __newElement(dom,'enttypds',ncgmp)
        enttyp.appendChild(enttypds)
        eanode.appendChild(detailed)        
    # __importMetadataFromDom has a trap for gdbName to avoid
    #   a failed attempt to import gdb-level metadata
    __importMetadataFromDom(dom, gdbName, gdbFolder)
    return dom

def __updateTable(domMR,fc,gdbFolder,titleSuffix,logFile,isAnno):
    #try to export metadata from fc
    if forceRebuild:
        domTb = copy.deepcopy(domMR)
    else:
        try:
            fcXML = __exportMD(fc, gdbFolder, 2)
            domTb = parse(fcXML)
        # but if this fails, use a copy of the master record
        except:
            domTb = copy.deepcopy(domMR)
    # adjust title
    domTb = __cleanTitle(domTb)
    domTb = __suffixTitle(domTb,titleSuffix)

    if entityDict.has_key(fc):
        descText = entityDict[fc]
        descSourceText = ncgmp
    else:
        descText = '**Need Description of '+fc+'**'
        descSourceText = ' '
        logFile.write('No description for feature class '+fc+'\n')

    eainfo = domTb.getElementsByTagName('eainfo')[0]

    if isAnno:
        # DELETE EXISTING CHILD NODES
        while len(eainfo.childNodes) > 0:
            eainfo.removeChild(eainfo.lastChild)
        # make new overview node
        overview = domTb.createElement('overview')
        eaover = __newElement(domTb,'eaover',descText)
        overview.appendChild(eaover)
        citationText = 'See ESRI documentation for structure of annotation feature classes.'
        eadetcit = __newElement(domTb,'eadetcit',citationText)
        overview.appendChild(eadetcit)
        # THEN ATTACH overview NODE
        eainfo.appendChild(overview)
    else:
        # check for e-a detailed node, add if necessary
        if len(eainfo.getElementsByTagName('detailed')) == 0:
            #add detailed/enttyp/enttypl nodes
            detailed = domTb.createElement('detailed')
            enttyp = domTb.createElement('enttyp')
            enttypl = __newElement(domTb,'enttypl',fc)
            enttypd = __newElement(domTb,'enttypd',descText)
            enttypds = __newElement(domTb,'enttypds',descSourceText)
            enttyp.appendChild(enttypl)
            enttyp.appendChild(enttypd)
            enttyp.appendChild(enttypds)
            detailed.appendChild(enttyp)
            eainfo.appendChild(detailed)
            
        ##check that each field has a corresponding attr node
        #get a list of the field names in the fc
        fldNameList = __fieldNameList(fc)    
        #get list of attributes in this metadata record
        #  we assume there eainfoNode has only one 'detailed' child
        attrlablNodes = eainfo.getElementsByTagName('attrlabl')
        attribs = []
        for nd in attrlablNodes:
            attribs.append(nd.firstChild.data)
        detailed = domTb.getElementsByTagName('detailed')[0]
        for fieldName in fldNameList:
            if not fieldName in attribs:
                attr = domTb.createElement('attr')
                attrlabl = __newElement(domTb,'attrlabl',fieldName)
                attr.appendChild(attrlabl)
                detailed.appendChild(attr)                
        #update the entity description and entity description source
        if entityDict.has_key(fc) or ( fc[0:2] == 'CS' and entityDict.has_key(fc[2:]) ):
            enttypl = domTb.getElementsByTagName('enttypl')
            if len(enttypl) > 0:
                enttyp = enttypl[0].parentNode
                # entity description node
                if fc[0:2] == 'CS':
                    descriptionText = entityDict[fc[2:]]
                else:
                    descriptionText = entityDict[fc]
                newEnttypd = __newElement(domTb,'enttypd',descriptionText)
                __appendOrReplace(enttyp,newEnttypd,'enttypd')
                # entity description source node
                newEnttypds = __newElement(domTb,'enttypds',ncgmp)
                __appendOrReplace(enttyp,newEnttypds,'enttypds')                           
        #update attribute descriptions and value domains
        domTb = __updateEntityAttributes(fc, fldNameList, domTb, logFile)
    
    #once we're done editing the xml, import it back into the 
    #embedded feature class metadata
    __importMetadataFromDom(domTb, fc, gdbFolder)
    addMsgAndPrint(gp, '\t' * 3 + 'Metadata for '+ fc+ ' has been updated')
    return domTb

def __appendOrReplace(rootNode,newNode,nodeTag):
    if len(rootNode.getElementsByTagName(nodeTag)) == 0:
        rootNode.appendChild(newNode)
    else:
        rootNode.replaceChild(newNode,rootNode.getElementsByTagName(nodeTag)[0])

def __updateFdsDom(domFD,domFC,fc,fDS,logFile):
    """Updates feature dataset metadata with detailed Entity Attribute
       for a feature class"""
    #get the 'detailed' node for this feature class from the 
    #feature class metadata. We assume there is only one such node
    fcNodes = domFC.getElementsByTagName('enttypl')
    ovNodes = domFC.getElementsByTagName('overview')
    if len(fcNodes) > 0:
        for enttypl in fcNodes:
            if enttypl.firstChild.data == fc:
                detailedFC = enttypl.parentNode.parentNode
    else:  # maybe this FC has an overview description, w/o details
        if len(ovNodes) > 0:  # make a detailed node
            detailedFC = domFC.createElement('detailed')
            enttyp = domFC.createElement('enttyp')
            enttypl = __newElement(domFC,'enttypl',fc)
            if entityDict.has_key(fc):
                enttypdText = entityDict[fc]
                enttypds = __newElement(domFC,'enttypds',ncgmp)
            else:
                enttypdText = '**Need description of '+fc+'**'
                logFile.write('No description for feature class '+fc+' in metadata for feature dataset '+fDS+'\n')
                enttypds = __newElement(domFC,'enttypds','')
            enttypd = __newElement(domFC,'enttypd',enttypdText)
            enttyp.appendChild(enttypl)
            enttyp.appendChild(enttypd)
            enttyp.appendChild(enttypds)
            detailedFC.appendChild(enttyp)
            
    #try to get the 'detailed' node for this feature class from the
    #feature dataset XML
    foundFcNode = False
    fcNodes = domFD.getElementsByTagName('enttypl')
    for enttypl in fcNodes:
        if enttypl.firstChild.data == fc:
            detailedFD = enttypl.parentNode.parentNode
            detailedParent = detailedFD.parentNode
            foundFcNode = True
    # if we didn't find FcNode
    if not foundFcNode:
        nodeList = domFD.getElementsByTagName('eainfo')
        detailedParent = nodeList[0]
        detailedParent.appendChild(detailedFC)
    else:
        detailedParent.replaceChild(detailedFC, detailedFD)    
    addMsgAndPrint(gp, '\t' * 3+ 'Metadata for '+ fc +' has been added to')
    addMsgAndPrint(gp, '\t' * 4 + fDS + ' metadata')

def __importMetadataFromDom(dom, objName, gdbFolder):
    #write DOM to XML, then import XML into geodatabase object
    dom.saveXML
    xmlName = os.path.join(gdbFolder, objName+'.xml')
    if os.path.exists(xmlName):
        os.remove(xmlName)
    outf = open(xmlName, 'w')
    dom.writexml(outf)
    outf.close()
    target = os.path.join(gdbPath,objName)
    if objName <> os.path.basename(gdbPath):
        gp.MetadataImporter_conversion(xmlName, target)
        if not saveXmlFiles:
            os.remove(xmlName)

def __fieldNameList(fc):
    """Returns a list of field names from Field.name in gp.ListFields"""
    fldList = gp.listfields(fc)
    nameList = []
    for fld in fldList:
        if not fld.name in ('OBJECTID', 'SHAPE','Shape', 'Shape_Length', 'Shape_Area'):
            nameList.append(fld.name)    
    return nameList
    
def __parentFolder(gdbPath):
    """Return the parent folder path of a geodatabase"""
    pFolder = gdbPath
    while not gp.describe(pFolder).datatype =='Folder':
        pFolder = os.path.split(pFolder)[0]   
    return pFolder

def __findInlineRef(sourceID):
    """finds the Inline reference for each DataSource_ID"""
    query = '"DataSources_ID" = \'' + sourceID + '\''
    rows = gp.searchcursor(dataSources, query)
    row = rows.next()
    if not row is None:
        #return row.Inline
        return row.Source
    else:
        return ""
    
def __updateEntityAttributes(fc, fldList, dom, logFile):
    """For each attribute (field) in fldList,
        adds attribute definition and definition source,
        classifies as range domain, unrepresentable-value domain or enumerated-value domain, and 
            for range domains, adds rangemin, rangemax, and units;
            for unrepresentable value domains, adds unrepresentable value statement; 
            for enumerated value domains:
            1) Finds all controlled-vocabulary fields in the table sent to it
            2) Builds a set of unique terms in each field, ie, the domain
            3) Matches each domain value to an entry in the glossary
            4) Builds a dictionary of term:(definition, source) items
            5) Takes the dictionary items and put them into the metadata
              document as Attribute_Domain_Values
        Field MapUnit in table DescriptionOfMapUnits is treated as a special-case.
        """
    cantfindTerm = []
    cantfindValue = []
    for fld in fldList:      
        addMsgAndPrint(gp, '\t' * 3+ 'Field: '+ fld)
        # if is _ID field or if field definition is available, update definition
        if fld.find('_ID') > -1 or attribDict.has_key(fld):
            dom = __updateAttrDef(fld,dom)
        else:
            cantfindTerm.append(fld)

        #if this is an _ID field
        if fld.find('_ID') > -1:
            dom = __updateUdom(fld,dom,unrepresentableDomainDict['_ID'])

        #if this is another unrepresentable-domain field
        if unrepresentableDomainDict.has_key(fld):
            dom = __updateUdom(fld,dom,unrepresentableDomainDict[fld])
            
        #if this is a defined range-domain field
        elif rangeDomainDict.has_key(fld):
            dom = __updateRdom(fld,dom)

        #if this is MapUnit in DMU
        elif fld == 'MapUnit' and fc == 'DescriptionOfMapUnits':
            dom = __updateUdom(fld,dom,unrepresentableDomainDict['default'])
            
        #if this is a defined Enumerated Value Domain field
        elif fld in enumeratedValueDomainFieldList:
            valList = []
            
            #create a search cursor on the field
            rows = gp.searchcursor(fc,'','', fld)
            row = rows.next()           
            #collect all values/terms in that field
            while row:
                if not row.GetValue(fld) is None:
                    valList.append(row.GetValue(fld))
                row = rows.next()            
            #uniquify the list by converting it to a set object
            valList = set(valList)
            
            #create an empty dictionary object to hold the matches between the unique terms
            #and their definitions (grabbed from the glossary)
            defs = {}
            
            #for each unique term, try to create a search cursor of just one record where the term
            #matchs a Term field value from the glossary
            if fld == 'MapUnit' and fc <> 'DescriptionOfMapUnits':
                for t in valList:
                    query = '"MapUnit" = \'' + t + '\''
                    rows = gp.searchcursor(DMU, query)
                    row = rows.next()
                    #if the searchcursor contains a row
                    if row:
                        #create an entry in the dictionary of term:[definition, source] key:value pairs
                        #this is how we will enumerate through the enumerated_domain section
                        defs[t] = []
                        defs[t].append(row.FullName.encode('utf_8'))
                        defs[t].append('')
                    else:
                        cantfindValue.append([fld,t])
            elif fld.find('SourceID') > -1:  # is a source field
                for t in valList:
                    query = '"DataSources_ID" = \'' + t + '\''
                    rows = gp.searchcursor(dataSources, query)
                    row = rows.next()
                    #if the searchcursor contains a row
                    if row:
                        #create an entry in the dictionary of term:[definition, source] key:value pairs
                        #this is how we will enumerate through the enumerated_domain section
                        defs[t] = []
                        defs[t].append(row.Source.encode('utf_8'))
                        defs[t].append('')
                    else:
                        cantfindValue.append([fld,t])
            else:
                for t in valList:
                    query = '"Term" = \'' + t + '\''
                    rows = gp.searchcursor(gloss, query)
                    row = rows.next()
                    #if the searchcursor contains a row
                    if row:
                        #create an entry in the dictionary of term:[definition, source] key:value pairs
                        #this is how we will enumerate through the enumerated_domain section
                        defs[t] = []
                        defs[t].append(row.Definition.encode('utf_8'))
                        defs[t].append(__findInlineRef(row.DefinitionSourceID).encode('utf_8'))
                    else:
                        cantfindValue.append([fld,t])
            dom = __updateEdom(fld, defs, dom)

        else:  #presumed to be an unrepresentable domain
            dom = __updateUdom(fld,dom,unrepresentableDomainDict['default'])

    if len(cantfindValue) > 0:
        addMsgAndPrint(gp, '\t' * 4+ 'Cannot find definition(s) for these term(s):')
        logFile.write('Missing enumerated-domain values\n')
        logFile.write('  ENTITY     TERM     VALUE\n')
        for term in cantfindValue:
            addMsgAndPrint(gp, '\t' * 5+ term[1])
            logFile.write('  '+fc+'  '+term[0]+' **'+term[1]+'**\n')
    if len(cantfindTerm) > 0:
        logFile.write('Missing terms\n')
        logFile.write('  ENTITY     TERM\n')
        for term in cantfindTerm:
            logFile.write('  '+fc + '  '+term+'\n')
    return dom

def __updateAttrDef(fld,dom):
    ##element tag names are
    ## attr             = Attribute
    ## attrlabl         = Attribute_Label
    ## attrdef          = Attribute_Definition
    ## attrdefs         = Attribute_Definition_Source
    labelNodes = dom.getElementsByTagName('attrlabl')
    for attrlabl in labelNodes:
        if attrlabl.firstChild.data == fld:
            attr = attrlabl.parentNode       
            if fld.find('_ID') > -1:
                # substitute generic _ID field for specific
                attrdefText = attribDict['_ID']
            else:
                attrdefText = attribDict[fld]
            attrdef = __newElement(dom,'attrdef',attrdefText)
            __appendOrReplace(attr,attrdef,'attrdef')
            attrdefs = __newElement(dom,'attrdefs',ncgmp)
            __appendOrReplace(attr,attrdefs,'attrdefs')
    return dom

def __updateUdom(fld,dom,udomTextString):
    labelNodes = dom.getElementsByTagName('attrlabl')
    for attrlabl in labelNodes:
        if attrlabl.firstChild.data == fld:
            attr = attrlabl.parentNode
            attrdomv = dom.createElement('attrdomv')
            udom = __newElement(dom,'udom',udomTextString)
            attrdomv.appendChild(udom)
            __appendOrReplace(attr,attrdomv,'attrdomv')
    return dom

def __updateRdom(fld,dom):
    labelNodes = dom.getElementsByTagName('attrlabl')
    for attrlabl in labelNodes:
        if attrlabl.firstChild.data == fld:
            attr = attrlabl.parentNode
            attrdomv = dom.createElement('attrdomv')
            rdom = dom.createElement('rdom')
            rdommin = __newElement(dom,'rdommin',rangeDomainDict[fld][0])
            rdom.appendChild(rdommin)
            rdommax = __newElement(dom,'rdommax',rangeDomainDict[fld][1])
            rdom.appendChild(rdommax)
            attrunit = __newElement(dom,'attrunit',rangeDomainDict[fld][2])
            rdom.appendChild(attrunit)
            attrdomv.appendChild(rdom)
            __appendOrReplace(attr,attrdomv,'attrdomv')
            return dom

def __updateEdom(fld, defs, dom):
    ##element tag names are
    ## attr             = Attribute
    ## attrdomv         = Attribute_Domain_Values
    ## edom             = Enumerated_Domain
    ## edomv            = Enumerated_Domain_Value
    ## edomd            = Enumerated_Domain_Definition
    ## edomvds          = Enumerated_Domain_Value_Definition_Source
    labelNodes = dom.getElementsByTagName('attrlabl')
    for attrlabl in labelNodes:
        if attrlabl.firstChild.data == fld:
            attr = attrlabl.parentNode
            attrdomv = dom.createElement('attrdomv')
            for k in defs.iteritems():
                edom = dom.createElement('edom')
                edomv = __newElement(dom,'edomv',k[0])
                edomvd = __newElement(dom,'edomvd',k[1][0])
                edom.appendChild(edomv)
                edom.appendChild(edomvd)
                if len(k[1][1]) > 0:
                    edomvds = __newElement(dom,'edomvds',k[1][1])
                    edom.appendChild(edomvds)                                
                attrdomv.appendChild(edom)
            __appendOrReplace(attr,attrdomv,'attrdomv')
    return dom

#################################################################

#create global geoprocessing object
gp = arcgisscripting.create(9.3)

#argument line" uncomment for use as tool
gdbPath = os.path.abspath(sys.argv[1])
addMsgAndPrint(gp,gdbPath)
if len(sys.argv) > 2:
    if sys.argv[2].lower() == 'true':
        saveXmlFiles = True
    elif sys.argv[2].lower() == 'false':
        saveXmlFiles = False
    else:
        addMsgAndPrint(gp,  '2nd argument (saveXmlFiles) should be True or False')
if len(sys.argv) > 3:
    if sys.argv[2].lower() == 'true':
        forceRebuild = True
    elif sys.argv[2].lower() == 'false':
        forceRebuild = False
    else:
        addMsgAndPrint(gp,  '3rd argument (forceRebuild) should be True or False')
       

#paths to some tables in the gdb
gloss = os.path.join(gdbPath, 'Glossary')
dataSources = os.path.join(gdbPath, 'DataSources')
DMU = os.path.join(gdbPath, 'DescriptionOfMapUnits')

addMsgAndPrint(gp,  versionString)
addMsgAndPrint(gp,  'Geodatabase: '+ gdbPath+'\n')
processMD(gdbPath)

#sys.stdout = save_stdout

## Still to be done:
#     Add definitions for attributes of GeologicEvents, ExtendedAttributes, StandardLithology
#     Write code to trap CrossSectionA <B> <C> etc and handle subsidiary feature classes

